# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/04_properties.ipynb (unless otherwise specified).

__all__ = ['generate_evolvers', 'generate_option_files', 'additional_properties', 'write_to_file',
           'structure_parameters', 'configurable_parameters', 'description', 'fixed_parameters']

# Cell
import os

# Cell
def generate_evolvers(iters, envs=None, collection=None, configs=None, properties=None, varieties=None):
    for env in envs:
        if not os.path.isdir(env):
            os.mkdir(env)

    for env in envs:
        num_actions = varieties[env]['num_actions']
        nevals = varieties[env]['nevals']
        archs = varieties[env]['archs']
        for arch in archs:
            key = '_'.join((env,arch['name']))
            print(key)
            config = configs[key]
            generate_option_files(iters, env, num_actions, arch, config, nevals, properties, collection)


# Cell
def generate_option_files(iters, env, num_actions, arch, config, nevals, properties, collection):
    arch_name = arch['name']
    inputs_names = arch['inputs_names']
    ppars = ''

    collectors=collection[env]['arch'][arch_name]['collectors']
    responses=collection[env]['arch'][arch_name]['responses']
    weights=collection[env]['arch'][arch_name]['weights']
    structs=collection[env]['arch'][arch_name]['structs']


    for collector in collectors:
        for response in responses:
            for weight in weights:
                for struct in structs:
                    desc, filename = description(collector,response, weight, struct, arch_name)
                    fpars = fixed_parameters(env, arch, num_actions)
                    cpars = configurable_parameters(env, config, collector, response, nevals)
                    ppars = additional_properties(properties, response, collector)
                    spars = structure_parameters(collector,response, weight, struct, arch_name)
                    display = f'### Display\n\ninputs_names = {inputs_names}\n'

                    text = '\n'.join((desc, fpars, cpars, ppars, spars, display))
                    filepath = f'{env}{os.sep}{filename}.properties'
                    write_to_file(filepath, text)
                    cmd = f'python run-dynamic-evolver-multi.py {filepath} -i {iters}'
                    print(cmd, end='\n')

# Cell
def additional_properties(properties, response, collector):
    ppars = ''

    if len(properties)>0:
        ppars='### Additional properties\n\n'

    ctr = 1
    for  prop in properties:
        value = properties[prop]
        if response == 'SmoothError' and prop == 'error:smooth_factor':
            propstr = f'property{ctr} = {prop},{value}'
            ppars = ''.join((ppars, propstr, '\n'))
            ctr+=1

        if collector == 'ReferencedInputsError' and  prop == 'error:referenced_inputs':
            propstr = f'property{ctr} = {prop},{value}'
            ppars = ''.join((ppars, propstr, '\n'))
            ctr+=1


    return ppars

# Cell
def write_to_file(file, text):
     f = open(file, "w")
     f.write(text)
     f.close()

# Cell
def structure_parameters(collector,response, weight, struct, arch):
    header = '### Structure\n\n# modes - pattern of nodes at particular levels, zero, n, top and zerotop\n# the mode numbers refer to# 3 - per:ws, ref:ws, com:sub, out:ws\n# 4 - per:ws, ref:con, com:sub, out:ws\n# 5 - per:ws, ref:con, com:sub, out:smws\n# 6 - per:ws, ref:ws, com:sub, out:smws\n'
    modes = [3, 3, 4, 4]
    if struct == 'SmoothWeightedSum':
        modes = [6, 6, 5, 5]

    mstr = f'modes = {modes}'
    type_num = 1
    types = ''
    if weight == 'Floats' or weight == 'AllFloats':
        types = ''.join((types, f'type{type_num} = [LevelKey.ZERO, perception, Float]\n'))
        type_num += 1
        types = ''.join((types, f'type{type_num} = [LevelKey.N, perception, Float]\n'))
        type_num += 1
        types = ''.join((types, f'type{type_num} = [LevelKey.TOP, perception, Float]\n'))
        type_num += 1

    if weight == 'AllFloats':
        types = ''.join((types, f'type{type_num} = [LevelKey.ZERO, action, Float]\n'))
        type_num += 1

    if struct == 'SmoothWeightedSum':
        types = ''.join((types, f'type{type_num} = [LevelKey.ZERO, output, Smooth]\n'))
        type_num += 1
        types = ''.join((types, f'type{type_num} = [LevelKey.N, output, Smooth]\n'))
        type_num += 1
        types = ''.join((types, f'type{type_num} = [LevelKey.TOP, output, Smooth]\n'))
        type_num += 1



    rtn = '\n'.join((header, mstr, types))
    return rtn

# Cell
def configurable_parameters(env, config, collector, response, nevals):
    header = ''.join(("### Configurable parameters\n\n# Randomisation seed to reproduce results\n# Size of population\n",
                     "# Number of generations\n# Probability that an attribute will be mutated\n# Probability that the structure will be mutated\n",
                     "# Number of runs of environment\n# Lower limit of float values\n# Upper limit of float values\n",
                     "# Initial limit of levels\n# Initial limit of columns\n# Lower limit of levels\n# Lower limit of columns\n",
                     "# Limit of error on which to terminate individual evaluation\n# Probability for crossover\n# Probability for mutating an individual\n# Number of times the evaulation is run (with different random seeds)\n# Type of errors collected\n# Error function\n\n"))

    text = ''
    for key in config.keys():
        value = config[key]
        text = ''.join((text, key, ' = ', f'{value}', '\n'))

    text = ''.join((header, text, f'nevals = {nevals}\nerror_collector = {collector}\nerror_response = {response}\n'))

    #f'seed = {seed}\nPOPULATION_SIZE = {POPULATION_SIZE}\nMAX_GENERATIONS = {MAX_GENERATIONS}\nattr_mut_pb={attr_mut_pb}\nstructurepb={structurepb}\nruns={runs}\nlower_float = {lower_float}\nupper_float = {upper_float}\nlevels_limit = {levels_limit}\ncolumns_limit = {columns_limit}\nerror_limit = {error_limit}\np_crossover = {p_crossover}\np_mutation = {p_mutation}\nnevals = {nevals}\nerror_collector = {error_collector}\nerror_response = {error_response}\n'
    return text


# Cell
def description(collector,response, weight, struct, arch):
    filename = '-'.join((arch, collector,response, weight, struct))
    rtn = ''.join(('\n### Description:\n\n','desc = ', filename,'\n'))
    return rtn, filename

# Cell
def fixed_parameters(env, option, num_actions):
    top_inputs = None
    inputs=option['inputs']
    if 'top_inputs' in option.keys():
        top_inputs=option['top_inputs']
    references=option['references']

    header = '### Fixed parameters\n\n# List of input indexes from environment\n# List of reference values\n# Number of actions\n\n'
    text1 = f'env = {env}\ninputs = {inputs}\n'
    if top_inputs == None:
        text2=''
    else:
        text2 = f'top_inputs = {top_inputs}\n'
    text3 = f'references = {references}\nnum_actions = {num_actions}\n'
    return ''.join((header,text1,text2,text3))




# Cell
import time
import warnings
warnings.simplefilter(action='ignore', category=UserWarning)
from .evolvers import evolve_from_properties_file

# Cell
import time
import warnings
warnings.simplefilter(action='ignore', category=UserWarning)
from .evolvers import evolve_from_properties_file

# Cell
import time
import warnings
import os
warnings.simplefilter(action='ignore', category=UserWarning)
from .evolvers import evolve_from_properties_file

# Cell
import time
import warnings
import os
warnings.simplefilter(action='ignore', category=UserWarning)
from .evolvers import evolve_from_properties_file

# Cell
import time
import warnings
import os
warnings.simplefilter(action='ignore', category=UserWarning)
from .evolvers import evolve_from_properties_file


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/01_evolvers.ipynb (unless otherwise specified).

__all__ = ['BaseEvolver', 'EvolverWrapper', 'SigmoidEvolver', 'StandardEvolver', 'DynamicEvolver',
           'check_hash_file_exists', 'evolve_from_properties_file']

# Cell
from jproperties import Properties
import random
import numpy as np
import math
import time
import enum
import networkx
import os
import socket
import hashlib
from networkx.drawing.nx_agraph import graphviz_layout
import matplotlib.pyplot as plt
import matplotlib.animation as animation
from matplotlib import style
from deap import algorithms
from abc import ABC, abstractmethod
from deap import base
from deap import creator
from deap import tools
from deap.tools import History
#from scoop import futures
import multiprocessing
from datetime import datetime

from .structure import StructureDefinition
from pct.putils import sigmoid
from .configs import ProportionalConfiguration
from .configs import BaseConfiguration
from .configs import DynamicConfiguration
from .configs import DynamicConfigurationStructure
from .configs import ConfigurationStructure
from pct.putils import stringIntListToListOfInts
from pct.putils import stringFloatListToListOfFloats
from pct.putils import stringListToListOfStrings
from pct.putils import listNumsToString
from pct.putils import get_drive
from pct.architectures import ProportionalArchitecture
from pct.architectures import DynamicArchitecture
from pct.structure  import ArchitectureStructure
from pct.nodes import ControlUnitIndices
from pct.architectures import LevelKey
from pct.environments import VelocityModel
from pct.environments import OpenAIGym
from .structure import BinaryOnes
from pct.errors import RootSumSquaredError
from pct.errors import TotalError
from pct.errors import BaseErrorCollector
from pct.putils import get_drive
from pct.environments import EnvironmentFactory
from pct.architectures import load_properties

# Cell
class BaseEvolver(ABC):
    "Base class of an evolver. This class is not used direclty by developers, but defines the interface common to all."
    def __init__(self):
        self.mutate_structure_sum=0
        self.mutate_sum=0

    def set_toolbox(self, toolbox):
        self.toolbox=toolbox

    def ensure_non_zero(self, item, index):
        if np.sum(item[index])==0:
            length = len(item[index])
            ind = random.randint(0, length-1)
            item[index][ind]=1

    def reset_mutate_structure_sum(self):
        sum = self.mutate_structure_sum
        self.mutate_structure_sum = 0
        return sum

    def reset_mutate_sum(self):
        sum = self.mutate_sum
        self.mutate_sum=0
        return sum

    def set_gen(self, gen):
        self.gen=gen

    def get_level_key(self, level, levels):
        key = LevelKey.N
        if level == 0:
            key = LevelKey.ZERO
        if level == levels-1:
            key = LevelKey.TOP

        return key

    @abstractmethod
    def create(self, cls):
        return cls(new_individial)

    @abstractmethod
    def evaluate(self, individual):
        return score,

    @abstractmethod
    def mate(self, ind1, ind2):
        return ind1, ind2

    @abstractmethod
    def mutate(self, ind, indpb=0.05):
        return ind,

    @abstractmethod
    def create_plot(self):
        return 0

    @abstractmethod
    def animate(self, epoch):
        pass

# Cell
class EvolverWrapper():

    def __init__(self, evolver, pop_size=25, p_crossover = 0.9, p_mutation = 0.1, display_env=False,
                 select={'selection_type':'tournament', 'tournsize':None}, parallel=False,
                 save_arch_gen=False, save_arch_all=False, **cargs):
        self.timing_sum=0
        self.save_arch_gen=save_arch_gen
        self.save_arch_all=save_arch_all
        self.display_env=display_env
        self.evolver=evolver
        self.pop_size=pop_size
        self.p_crossover = p_crossover   # probability for crossover
        self.p_mutation = p_mutation   # probability for mutating an individual
        if hasattr(creator, 'FitnessMin'):
            del creator.FitnessMin
            del creator.Individual
        creator.create("FitnessMin", base.Fitness, weights=(-1.0,))
        creator.create("Individual", list, fitness=creator.FitnessMin)
        # Structure initializers
        self.toolbox = base.Toolbox()

        if parallel:
            pool = multiprocessing.Pool()
            self.toolbox.register("map", pool.map)
            #self.toolbox.register("map", futures.map)

        self.toolbox.register("individual", self.create(), creator.Individual)
        self.toolbox.register("population", tools.initRepeat, list, self.toolbox.individual)

        self.toolbox.register("evaluate", self.evaluate())
        self.toolbox.register("mate", self.mate())
        self.toolbox.register("mutate", self.mutate())

        if select['selection_type'] == 'tournament':
            if 'tournsize' in select.keys():
                tournsize=select['tournsize']
            else:
                tournsize=None
            if tournsize==None:
                tournsize=(int)(pop_size/4)
            self.toolbox.register("select", tools.selTournament, tournsize=tournsize)

        if select['selection_type'] ==  'best':
            self.toolbox.register("select", tools.selBest)

        genealogy = History()
        self.toolbox.decorate("mate", genealogy.decorator)
        self.toolbox.decorate("mutate", genealogy.decorator)
        self.hof = tools.HallOfFame(1)

        #self.pop = self.toolbox.population(n=pop_size)
        #history.update(self.pop)

        self.genealogy=genealogy

        self.stats = tools.Statistics(lambda ind: ind.fitness.values)
        self.stats.register("mean", np.mean)
        self.stats.register("min", np.min)
        self.stats.register("max", np.max)
        self.stats.register("std", np.std)
        self.evolver.set_toolbox(self.toolbox)
        self.stats_history={}
        self.stats_history['min'] = []

    def create(self):
        return self.evolver.create

    def evaluate(self):
        return self.evolver.evaluate

    def mate(self):
        return self.evolver.mate

    def mutate(self):
        return self.evolver.mutate

    def create_population(self):
        self.pop = self.toolbox.population(n=self.pop_size)
        return self.pop

    def collect_stats(self, gen, index, logbook, timeperpop, verbose, log):
        min = logbook[index]['min']
        mean = logbook[index]['mean']
        max = logbook[index]['max']
        #std = logbook[index]['std']
        log_string=''
        self.stats_history['min'].append(min)

        if index > 0:
            mut = 100*self.evolver.reset_mutate_sum()/self.pop_size
            muts = 100*self.evolver.reset_mutate_structure_sum()/self.pop_size

        if verbose>0 or log:
            if index > 0:
                logs = f'{gen:4} {len(self.pop):4} {min:8.3f} {mean:10.3f} {max:10.3f}  {mut:3.0f}% {muts:3.0f}%   {timeperpop:4.3f}'
            else:
                logs = f'{gen:4} {len(self.pop):4} {min:8.3f} {mean:10.3f} {max:10.3f} \n'

        if log:
            log_string = ''.join(('# ', logs))
        if verbose>0:
            print(logs, end = '')

        return log_string

    def run(self, gens=25, verbose=False, deap_verbose=False, log=False):
        log_string = ''

        if self.save_arch_all:
            self.evolver.set_save_arch_all(self.save_arch_all)
            self.evolver.set_gen(0)

        self.pop = self.toolbox.population(n=self.pop_size)
        self.genealogy.update(self.pop)

        if verbose>0:
            logs = 'gen   pop   min       mean      max        mut  muts  timing'
            print(logs)
        if log:
            logs = 'gen   pop   min       mean      max        mut  muts  timing'
            log_string = ''.join((log_string, '# ', logs, '\n'))

        self.best_of_gens=[]
        top_ind = tools.selBest(self.pop, k=1)[0]
        self.best_of_gens.append(top_ind)

        for gen in range(1, gens+1, 1):
            self.evolver.set_gen(gen)
            tic = time.perf_counter()
            self.pop, logbook = algorithms.eaSimple(self.pop, self.toolbox, cxpb=self.p_crossover,
                    mutpb=self.p_mutation,  ngen=1, halloffame=self.hof, stats=self.stats, verbose=deap_verbose)
            toc = time.perf_counter()
            timeperpop = (toc-tic)/self.pop_size
            self.timing_sum += timeperpop
            self.evolver.member=0
            if gen == 1:
                log_stats = self.collect_stats( 0, 0, logbook, timeperpop, verbose, log)
                if log:
                    log_string = ''.join((log_string, log_stats))

            log_stats = self.collect_stats( gen, 1, logbook, timeperpop, verbose, log)

            if verbose>2:
                print('pop ***')
                for pop in self.pop:
                    print(pop)

            top_ind = tools.selBest(self.pop, k=1)[0]
            if verbose>1:
                print ( f'[{top_ind}]')
            else:
                if verbose>0:
                    print()
                if log:
                    log_string = ''.join((log_string, log_stats, '\n'))

            self.best_of_gens.append(top_ind)
            if self.display_env:
                if hasattr(self.evolver, 'env'):
                    if isinstance(self.evolver.env, OpenAIGym):
                        self.evolver.env.set_render(True)
                        if self.save_arch_gen:
                            self.evolver.fig_file = f'fig{gen:03}.png'
                        if gen == 1:
                            print('Displaying gen 0')
                            self.evolver.env.set_render(True)
                            self.evolver.evaluate(self.best_of_gens[0])
                            print('>> Press enter to continue:')
                            input()
                            print('Displaying gen 0')
                            self.evolver.env.set_render(True)
                            self.evolver.evaluate(self.best_of_gens[0])
                            self.evolver.env.set_render(True)
                        print(f'Displaying gen {gen}')
                        self.evolver.evaluate(top_ind)
                        self.evolver.fig_file = None
                        self.evolver.env.set_render(False)
                        print(top_ind)

        return self.timing_sum/gens, log_string

    def draw_tree(self, node_size=500, with_labels=True):
        graph = networkx.DiGraph(self.genealogy.genealogy_tree)
        graph = graph.reverse()     # Make the graph top-down
        colors = [self.toolbox.evaluate(self.genealogy.genealogy_history[i])[0] for i in graph]
        pos=graphviz_layout(graph, prog="dot")
        figsize=(14,20)
        plt.figure(figsize=figsize)
        networkx.draw(graph, pos=pos, node_color=colors, node_size=node_size,  with_labels=with_labels)

    def best(self):
        return self.hof[0]

    def best_score(self):
        return self.hof[0].fitness.values[0]

    def get_best_of_gens(self):
        return self.best_of_gens

    def create_stats_plot(self, data):
        first_key = list(data.keys())[0]
        self.x = np.linspace(0, len(data[first_key])-1, len(data[first_key]))
        style.use('fivethirtyeight')
        self.fig = plt.figure(figsize=(10,5))
        self.ax1 = self.fig.add_subplot(1,1,1)
        plt.title('Stats')
        self.ax1.grid(True)

        for datum in data:
            self.ax1.plot(self.x, data[datum],  c='r')

        return self.fig


# Cell
class SigmoidEvolver(BaseEvolver):
    "A class to evolve a sigmoid function defined by two parameters the range and scale (slope)."
    'Parameters:'
    'range_limit - the initial limit of the range for an individual'
    'scale_limit - the initial limit of the scale for an individual'
    'alpha - for crossover, the extent to which the new values can be drawn on both side of the parents’ attributes'
    'mu - for mutation, mean for the gaussian mutation'
    'sigma - standard deviation for the gaussian mutation'
    'indpb - probability for each attribute to be mutated'
    'evaluate_factor - factor at which goal target is evaulated'
    'evaluate_divisor - division of goal at which half the goal is evaluated'

    def __init__(self, goal=7, range_limit=20, scale_limit=5, alpha=0.5, mu=0.1, sigma=0.25, indpb=0.2,
                 evaluate_factor=5, evaluate_divisor=5,  **cargs):
        super().__init__()
        self.goal=goal
        self.range_limit=range_limit
        self.scale_limit=scale_limit
        self.alpha=alpha
        self.mu=mu
        self.sigma=sigma
        self.evaluate_factor=evaluate_factor
        self.evaluate_divisor=evaluate_divisor
        self.indpb=indpb

    def create(self, cls):
        new_individial = []
        new_individial.append(random.uniform(0, self.range_limit))
        new_individial.append(random.uniform(0, self.scale_limit))

        return cls(new_individial)

    def evaluate(self, individual):

        shift = 0
        upper = 0
        lower = 0
        zero = 0

        goal = self.goal * self.evaluate_factor
        upper =shift + sigmoid(goal, individual[0], individual[1])
        lower = shift + sigmoid(self.goal/self.evaluate_divisor, individual[0], individual[1])
        zero = sigmoid(0, individual[0], individual[1])

        score = abs(self.goal - upper) + abs(self.goal/2 - lower) + shift - zero
        return score,

    def mate(self, ind1, ind2):
        ind1, ind2 = tools.cxBlend(ind1, ind2, self.alpha)
        if ind1[0]<0 or ind2[0]<0 :
            if ind1[0]<0 :
               ind1[0]=abs(ind1[0])
            if ind2[0]<0 :
               ind2[0]=abs(ind2[0])

        if ind1[1]<0 or ind2[1]<0 :
            if ind1[1]<0 :
               ind1[1]=abs(ind1[1])
            if ind2[1]<0 :
               ind2[1]=abs(ind2[1])

        return ind1, ind2

    def mutate(self, ind):

        mutant = ind
        #print('indpb',self.indpb)
        mutant, = tools.mutGaussian(ind, mu=self.mu, sigma=self.sigma, indpb=self.indpb)
        return mutant,


    def create_plot(self, best_of_gens):
        self.best_of_gens=best_of_gens
        self.x = np.linspace(-50, 50, 101)
        self.frames = len(best_of_gens)
        top_ind = self.best_of_gens[-1]
        self.ytarget = sigmoid(self.x, top_ind[0], top_ind[1])
        style.use('fivethirtyeight')
        self.fig = plt.figure(figsize=(10,5))
        self.ax1 = self.fig.add_subplot(1,1,1)
        plt.title('Target')
        self.ax1.grid(True)
        self.ax1.plot(self.x, self.ytarget,  c='r')
        return self.fig

    def animate(self, epoch):
        if epoch <= self.frames:
            top_ind = self.best_of_gens[epoch]
            y = sigmoid(self.x, top_ind[0], top_ind[1])
            self.ax1.clear()
            self.ax1.grid(True)
            self.ax1.plot(self.x, self.ytarget,  c='r')
            self.ax1.plot(self.x, y,  c='b', linewidth=2)

# Cell
class StandardEvolver(BaseEvolver):
    "A class to evolve PCT hierarchies."
    'Parameters:'
    'error_collector - the class for collecting the errors'
    'env - the environment simulation'
    'runs - the number of generations'
    'alpha - for crossover, the extent to which the new values can be drawn on both side of the parents’ attributes'
    'mu - for mutation, mean for the gaussian mutation'
    'sigma - standard deviation for the gaussian mutation'
    'inputs - the list of the indices of the perceptions from the environment'
    'num_actions - the number of actions in the environment'
    'references - the list of reference values'
    'levels_limit - the maximum number of levels in initial population individuals'
    'columns_limit - the maximum number of comumns in initial population individuals'
    'lower_float - the lower limit of float values for weights'
    'upper_float - the upper limit of float values for weights'
    'structurepb - the probability of a change in structure'
    'indpb - probability for each attribute to be mutated'

    def __init__(self, env=None, inputs=None, num_actions=1, references=[1], levels_limit=3,
                 columns_limit=3, alpha=0.5, mu=0.1, sigma=0.25, indpb=0.2, runs=100,
                 lower_float=-100, upper_float=100, error_response_type=None, error_collector_type=None,
                 error_limit=None, structurepb=0.1, debug=0, seed=None, hpct_verbose=False,
                 nevals=1, file= None, **cargs):
        super().__init__()
        if inputs==None:
            raise Exception('Inputs must be defined')
        self.nevals = nevals
        self.hpct_verbose=hpct_verbose
        self.error_collector_type=error_collector_type
        self.error_response_type=error_response_type
        self.error_limit=error_limit
        if error_collector_type == None:
            self.error_collector_type='TotalError'
        if error_response_type == None:
            self.error_response_type='RootSumSquaredError'
        if error_limit == None:
            self.error_limit=500
        self.env=env
        self.runs=runs
        self.alpha=alpha
        self.mu=mu
        self.sigma=sigma
        self.inputs=inputs
        self.num_actions=num_actions
        self.references=references
        self.levels_limit=levels_limit
        self.columns_limit=columns_limit
        self.lower_float=lower_float
        self.upper_float=upper_float
        self.structurepb=structurepb
        self.debug=debug
        self.indpb=indpb
        self.seed=seed
        random.seed(seed)
        if self.debug>0:
            print('Initial population')
        self.fig_file=file
        self.member=0
        self.save_arch_all=False


    def set_save_arch_all(self, save_arch_all):
        self.save_arch_all = save_arch_all

    def create(self, cls):
        grid=[]
        levels = random.randint(0, self.levels_limit)
        #print(levels)
        for level in range(levels-1):
            columns = random.randint(1, self.columns_limit)
            grid.append(columns)

        grid.append(len(self.references))

        #print(grid)
        new_individual = ProportionalConfiguration(num_inputs=len(self.inputs), num_actions=self.num_actions,
            grid=grid, references=self.references, lower_float=self.lower_float, upper_float=self.upper_float)
        new_individual()
        if self.debug>0:
            print(new_individual.to_raw())
        return cls(new_individual.to_raw())

    def evaluate(self, individual):

        if self.save_arch_all:
            self.fig_file = f'arch-{self.gen:04}-{self.member:04}.png'

        config = BaseConfiguration.from_raw(individual)

        score = 0

        for i in range(self.nevals):
            error_collector = BaseErrorCollector.collector(self.error_response_type, self.error_collector_type, self.error_limit)
            if self.seed != None:
                self.env.set_seed(self.seed)
            self.env.reset()
            pa = ProportionalArchitecture(config=config, env=self.env, input_indexes=self.inputs, error_collector=error_collector)
            pa()
            hpct = pa.get_hierarchy()
            if self.fig_file != None:
                hpct.draw(file=self.fig_file)

            hpct.run(steps=self.runs, verbose=self.hpct_verbose)
            score += hpct.get_error_collector().error()

        if self.debug>1:
            print(f'{self.gen:03} {self.member:03} score {score:5.3f}', config)

        self.member+=1

        return score,


    def datatype(self, data):
        print(data)
        type = None
        if isinstance(data[0], int):
            return "int"

        return type

    def process_crossover(self, child1item, child2item, index):
        if isinstance(child1item[index][0], int):
            child1item[index], child2item[index] = tools.cxUniform(child1item[index], child2item[index], 0.5)
            self.ensure_non_zero(child1item, index)
            self.ensure_non_zero(child2item, index)

        if isinstance(child1item[index][0], float):
            child1item[index], child2item[index] = tools.cxBlend(child1item[index], child2item[index], self.alpha)


    def mate(self, ind1, ind2):
        child1, child2 = [self.toolbox.clone(ind) for ind in (ind1, ind2)]

        child2levellen = len(child2)
        for level in range(len(child1)):
            if level > child2levellen-1:
                break
            for item in range(3):#len(child1[level])):
                #if item ==3: print(child1[level][item])
                #print('level', level, 'item',item)
                if isinstance(child1[level][item][0], list):
                    child2itemlen = len(child2[level][item])
                    for index in range(len(child1[level][item])):
                        #print(index, child2itemlen)
                        if index > child2itemlen-1:
                            break
                        self.process_crossover(child1[level][item], child2[level][item], index)
                else:
                    self.process_crossover(child1[level], child2[level], item)

        if self.debug>2:
            print('mate b4',ind1)
            print('mate b4',ind2)
            print('mate b5',child1)
            print('mate b5',child2)

        return child1, child2

    def mutate(self, ind):
        mutated = 0
        mutated_structure = 0
        mutant = self.toolbox.clone(ind)
        if self.debug>2:
            print('mut b4',mutant)
        for level in range(len(mutant)):
            for item in range(3):
                if isinstance(mutant[level][item][0], list):
                    for index in range(len(mutant[level][item])):
                        mutated += self.process_mutant(mutant[level][item], index)
                else:
                    mutated += self.process_mutant(mutant[level], item)

        if random.random() < self.structurepb:
            mutant, mutated_structure = self.mutate_structure(mutant)

        if mutated>0 or mutated_structure>0:
            self.mutate_sum += 1

        if mutated_structure>0:
            self.mutate_structure_sum += 1

        if self.debug>2:
            print('mut b5',mutant)
        return mutant,

    def mutate_structure(self, mutant):
        mutated_structure = 0
        raw=mutant
        #print('mutate_structure')
        #print(raw)
        orig_config=BaseConfiguration.from_raw(raw)
        cs = ConfigurationStructure()

        struct = cs.from_raw(raw)
        #print()
        #print(struct)

        choice = random.randint(1,4)
        #print('choice',choice)
        levels = len(struct['grid'])
        if choice == 1 :
            upper_levels_limit = levels-2
            if upper_levels_limit >= 0:
                level = random.randint(0, levels-2)
                num_nodes = random.randint(1, self.columns_limit)
                #print(level, num_nodes)
                struct, mutated_structure = cs.add_nodes(level=level, num_nodes=num_nodes)

        if choice == 2 :
            num_columns = random.randint(1, self.columns_limit)
            struct, mutated_structure = cs.add_level(num_columns=num_columns)

        if choice == 3 :
            struct, mutated_structure = cs.remove_level()

        if choice == 4 :
            upper_levels_limit = levels-2
            if upper_levels_limit >= 0:
                remove_level = random.randint(0, upper_levels_limit)
                upper_node_limit = struct['grid'][remove_level]-1
                if upper_node_limit > 1:
                    remove_nodes = random.randint(1, upper_node_limit)
                    #print(remove_level, remove_nodes)
                    struct, mutated_structure = cs.remove_nodes(level=remove_level, num_nodes=remove_nodes)

        #print(struct)


        pc=ProportionalConfiguration(lower_float=self.lower_float,upper_float=self.upper_float, **struct)
        config=pc()
        #print(config)


        if choice == 1 :
            BaseConfiguration.copy_data(orig_config, config)
        if choice == 2 :
            BaseConfiguration.copy_data_from_smaller(orig_config, config)
        if choice == 3 :
            if len(orig_config) != len(config):
                #print(orig_config)
                #print(config)
                BaseConfiguration.copy_data_from_larger(orig_config, config)
            else:
                config=orig_config
        if choice == 4 :
            BaseConfiguration.copy_data(orig_config, config)

        #print(config)

        mutant = creator.Individual(BaseConfiguration.dict_to_raw(config))

        return mutant, mutated_structure

    def process_mutant(self, mutant, index):
        mutated = 0
        if isinstance(mutant[index][0], int) and len(mutant[index])>1:
            orig = self.toolbox.clone(mutant[index])
            mutant[index], = tools.mutFlipBit(mutant[index], self.indpb)
            if np.sum(mutant[index]) == 0:
                pos = random.randint(0, len(mutant[index])-1)
                #print('pos', pos)
                #print(mutant[index])
                mutant[index][pos]=1
            if orig != mutant[index]:
                mutated = 1

        if isinstance(mutant[index][0], float):
            orig = self.toolbox.clone(mutant[index])
            mutant[index], = tools.mutGaussian(mutant[index],  self.mu, self.sigma, self.indpb)
            if orig != mutant[index]:
                mutated = 1

        return mutated

    """
    def create_stats_plot(self, data):
        first_key = list(data.keys())[0]
        self.x = np.linspace(0, len(data[first_key])-1, len(data[first_key]))
        style.use('fivethirtyeight')
        self.fig = plt.figure(figsize=(10,5))
        self.ax1 = self.fig.add_subplot(1,1,1)
        plt.title('Stats')
        self.ax1.grid(True)

        for datum in data:
            self.ax1.plot(self.x, data[datum],  c='r')

        return self.fig
    """

    def create_plot(self, best_of_gens):
        self.best_of_gens=best_of_gens
        self.x = np.linspace(0, self.runs-1, self.runs)
        self.frames = len(best_of_gens)
        top_ind = self.best_of_gens[-1]
        config = BaseConfiguration.from_raw(top_ind)
        if self.debug == 1:
            print(config)
        self.env.reset()
        pa = ProportionalArchitecture(config=config, env=self.env, input_indexes=self.inputs, history=True)
        pa()
        hpct = pa.get_hierarchy()
        hpct.run(steps=self.runs)
        self.ytarget =  hpct.get_node(0, 0).history.data['percoll']['PL0C0ws']
        style.use('fivethirtyeight')
        self.fig = plt.figure(figsize=(10,5))
        self.ax1 = self.fig.add_subplot(1,1,1)
        plt.title('Target')
        self.ax1.grid(True)
        self.ax1.plot(self.x, self.ytarget,  c='r')
        return self.fig

    def animate(self, epoch):
        if epoch <= self.frames:
            top_ind = self.best_of_gens[epoch]
            config = BaseConfiguration.from_raw(top_ind)
            if self.debug == 1:
                print(config)
            self.env.reset()
            pa = ProportionalArchitecture(config=config, env=self.env, input_indexes=self.inputs, history=True)
            pa()
            hpct = pa.get_hierarchy()
            hpct.run(steps=self.runs)
            y =  hpct.get_node(0, 0).history.data['percoll']['PL0C0ws']
            self.ax1.clear()
            self.ax1.grid(True)
            self.ax1.plot(self.x, self.ytarget,  c='r')
            self.ax1.plot(self.x, y,  c='b', linewidth=2)


# Cell
class DynamicEvolver(BaseEvolver):
    "A class to evolve PCT hierarchies from a dynamic structure defintion."
    def __init__(self, env=None, inputs=None, num_actions=1, references=[1], alpha=0.5,
                 mu=0.1, sigma=0.25, indpb=0.2, runs=100,
                 error_response_type=None, error_collector_type=None,
                 error_limit=None, structurepb=0.1, debug=0, seed=None, hpct_verbose=False,
                 nevals=1, file= None, structure=None, error_properties=None, **cargs):
        super().__init__()
        if inputs==None:
            raise Exception('Inputs must be defined')
        self.nevals = nevals
        self.hpct_verbose=hpct_verbose
        self.error_collector_type=error_collector_type
        self.error_response_type=error_response_type
        self.error_limit=error_limit
        if error_collector_type == None:
            self.error_collector_type='TotalError'
        if error_response_type == None:
            self.error_response_type='RootSumSquaredError'
        if error_limit == None:
            self.error_limit=500
        self.env=env
        self.runs=runs
        self.structurepb=structurepb
        self.inputs=inputs
        self.num_actions=num_actions
        self.references=references
        self.debug=debug
        self.seed=seed
        random.seed(seed)
        if self.debug>0:
            print('Initial population')
        self.fig_file=file
        self.member=0
        self.save_arch_all=False
        self.structure=structure
        self.error_properties = error_properties


    def set_save_arch_all(self, save_arch_all):
        self.save_arch_all = save_arch_all

    def create(self, cls):


        #print(grid)
        new_individual = DynamicConfiguration(num_inputs=len(self.inputs), num_actions=self.num_actions,
            references=self.references, structure=self.structure)
        new_individual()
        #for key in new_individual.config.keys():
        #    print(key, new_individual.config[key])
        if self.debug>0:
            print(new_individual.to_raw())
        return cls(new_individual.to_raw())



    def evaluate(self, individual):

        if self.save_arch_all:
            self.fig_file = f'arch-{self.gen:04}-{self.member:04}.png'

        config = BaseConfiguration.from_raw(individual)

        #print(config)
        score = 0

        top_inputs = top_inputs = self.structure.get_parameter('top_inputs')

        for i in range(self.nevals):
            error_collector = BaseErrorCollector.collector(self.error_response_type, self.error_collector_type,
                                                           self.error_limit, properties=self.error_properties)
            if self.seed != None:
                self.env.set_seed(self.seed+i)
            self.env.reset()
            if i > 0:
                self.env.set_render(False)

            pa = DynamicArchitecture(structure=self.structure.arch_structure, config=config, env=self.env,
                                     input_indexes=self.inputs, top_input_indexes=top_inputs, error_collector=error_collector)
            pa()
            hpct = pa.get_hierarchy()

            #hpct.summary()

            if self.fig_file != None:
                hpct.draw(file=self.fig_file)

            hpct.run(steps=self.runs, verbose=self.hpct_verbose)
            score +=  hpct.get_error_collector().error()

        score = score / self.nevals
        if self.debug>1:
            print(f'{self.gen:03} {self.member:03} score {score:5.3f}', config)

        self.member+=1

        return score,



    def mate(self, ind1, ind2):
        child1, child2 = [self.toolbox.clone(ind) for ind in (ind1, ind2)]
        #print('c1', child1)
        #print('c2', child2)

        child1levels = len(child1)
        child2levels = len(child2)

        # actions, at level 0
        level=0
        for item in range(len(child1[level][ControlUnitIndices.ACT_INDEX])):
            self.structure.mate_lists(LevelKey.ZERO , 'action',  child1[level][ControlUnitIndices.ACT_INDEX][item], child2[level][ControlUnitIndices.ACT_INDEX][item])

        # perceptions and outputs at all levels
        for level in range(min(child1levels,child2levels)):
            level_key = self.get_level_key(level, child1levels)
            for item in range(min(len(child1[level][ControlUnitIndices.PER_INDEX]), len(child2[level][ControlUnitIndices.PER_INDEX]))):
                self.structure.mate_lists(level_key, 'perception',  child1[level][ControlUnitIndices.PER_INDEX][item],child2[level][ControlUnitIndices.PER_INDEX][item])

            data_type,x = self.structure.get_type(level_key, 'perception')
            if(data_type == 'Binary'):
                self.structure.ensure_non_zero(level_key, 'perception', child1[level][ControlUnitIndices.PER_INDEX])
                self.structure.ensure_non_zero(level_key, 'perception', child2[level][ControlUnitIndices.PER_INDEX])

            self.structure.mate_lists(level_key, 'output',  child1[level][ControlUnitIndices.OUT_INDEX], child2[level][ControlUnitIndices.OUT_INDEX])

        # references at all levels except top
        for level in range(min(child1levels-1,child2levels-1)):
            level_key = self.get_level_key(level, child1levels)
            #print(level, child1levels, level_key)
            for item in range(min(len(child1[level][ControlUnitIndices.REF_INDEX]), len(child2[level][ControlUnitIndices.REF_INDEX]))):
                self.structure.mate_lists(level_key, 'reference',  child1[level][ControlUnitIndices.REF_INDEX][item],child2[level][ControlUnitIndices.REF_INDEX][item])


        if self.debug>2:
            print('mate b4',ind1)
            print('mate b4',ind2)
            print('mate b5',child1)
            print('mate b5',child2)

        return child1, child2

    def mutate(self, ind, indpb=0.0):
        mutated=False
        mutated_structure = 0
        mutant = self.toolbox.clone(ind)

        if self.debug>2:
            print('mut b4')#,mutant)
            for i in range(len(mutant)):
                print(i, mutant[i])

        levels = len(mutant)

        level=0
        for item in range(len(mutant[level][ControlUnitIndices.PER_INDEX])):
            self.structure.mutate_list(LevelKey.ZERO , 'perception',  mutant[level][ControlUnitIndices.PER_INDEX][item])
        self.structure.mutate_list(LevelKey.ZERO , 'output',  mutant[level][ControlUnitIndices.OUT_INDEX])

        for item in range(len(mutant[level][ControlUnitIndices.ACT_INDEX])):
            self.structure.mutate_list(LevelKey.ZERO , 'action',  mutant[level][ControlUnitIndices.ACT_INDEX][item])



        if levels >1:
            for item in range(len(mutant[level][ControlUnitIndices.REF_INDEX])):
                self.structure.mutate_list(LevelKey.ZERO , 'reference',  mutant[level][ControlUnitIndices.REF_INDEX][item])
            if levels > 2:
                for level in range(1, levels-2, 1):
                    for item in range(len(mutant[level][ControlUnitIndices.PER_INDEX])):
                        self.structure.mutate_list(LevelKey.N, 'perception',  mutant[level][ControlUnitIndices.PER_INDEX][item])

                    self.structure.mutate_list(LevelKey.N, 'output',  mutant[level][ControlUnitIndices.OUT_INDEX])

                    for item in range(len(mutant[level][ControlUnitIndices.REF_INDEX])):
                        self.structure.mutate_list(LevelKey.N, 'reference',  mutant[level][ControlUnitIndices.REF_INDEX][item])

            toplevel=levels-1

            if self.structure.get_config_parameter(LevelKey.TOP, 'perception', 'ones') == BinaryOnes.AT_LEAST_ONE and len(mutant[toplevel][ControlUnitIndices.PER_INDEX][0])==1:
                self.structure.mutate_binary_lists(LevelKey.TOP, 'perception',mutant[toplevel][ControlUnitIndices.PER_INDEX])
            else:
                for item in range(len(mutant[toplevel][ControlUnitIndices.PER_INDEX])):
                    self.structure.mutate_list(LevelKey.TOP, 'perception',  mutant[toplevel][ControlUnitIndices.PER_INDEX][item])

            self.structure.mutate_list(LevelKey.TOP, 'output',  mutant[toplevel][ControlUnitIndices.OUT_INDEX])

        if random.random() < self.structurepb:
            mutant, mutated_structure = self.mutate_structure(mutant)

        if ind != mutant:
            mutated=True

        if mutated or mutated_structure>0:
            self.mutate_sum += 1

        if mutated_structure>0:
            self.mutate_structure_sum += 1

        if self.debug>2:
            print('mut b5')#,mutant)
            for i in range(len(mutant)):
                print(i, mutant[i])

        return mutant,




    def mutate_structure(self, mutant):
        mutated_structure = 0
        raw=mutant
        #print('b4 mutate_structure')
        #for i in range(len(mutant)):
        #    print(i, mutant[i])

        orig_config=BaseConfiguration.from_raw(raw)
        cs = DynamicConfigurationStructure()

        struct = cs.from_raw(raw)
        #print()
        #print(struct)

        choice = random.randint(1,4)
        #print('choice',choice)
        levels = len(struct['grid'])
        if choice == 1 :
            upper_levels_limit = levels-2
            if upper_levels_limit >= 0:
                level = random.randint(0, levels-2)
                max_columns_limit = self.structure.get_parameter('columns_limit')
                columns = struct['grid'][level]
                if columns < max_columns_limit:
                    num_nodes = random.randint(1, max_columns_limit-columns)
                    struct, mutated_structure = cs.add_nodes(level=level, num_nodes=num_nodes)

        if choice == 2 :
            max_levels_limit = self.structure.get_parameter('levels_limit')
            if levels < max_levels_limit:
                num_columns = random.randint(1, self.structure.get_parameter('columns_limit'))
                struct, mutated_structure = cs.add_level(num_columns=num_columns)

        if choice == 3 :
            min_levels_limit = self.structure.get_parameter('min_levels_limit')
            if levels > min_levels_limit:
                struct, mutated_structure = cs.remove_level()

        if choice == 4 :
            upper_levels_limit = levels-2
            if upper_levels_limit >= 0:
                remove_level = random.randint(0, upper_levels_limit)
                upper_node_limit = struct['grid'][remove_level]-1
                min_columns_limit = self.structure.get_parameter('min_columns_limit')
                columns = struct['grid'][remove_level]
                if upper_node_limit > 1:
                    if columns > min_columns_limit:
                        remove_nodes = random.randint(1, upper_node_limit)
                        #print(remove_level, remove_nodes)
                        struct, mutated_structure = cs.remove_nodes(level=remove_level, num_nodes=remove_nodes)

        #print(struct)

        pc=DynamicConfiguration(inputs=self.inputs, structure=self.structure, **struct)
        #pc=DynamicConfiguration( structure=self.structure, **struct)
        config=pc()
        #print('oc',orig_config)
        #print('c ',config)

        if mutated_structure==0 or (choice == 3 and len(orig_config) == len(config)):
            config=orig_config
        else:
            for item in range(len(orig_config['level0'][ControlUnitIndices.ACT_INDEX])):
                self.structure.copy_data(LevelKey.ZERO, 'action',  orig_config['level0'][ControlUnitIndices.ACT_INDEX][item], config['level0'][ControlUnitIndices.ACT_INDEX][item])

            from_top_level =  max(BaseConfiguration.top_level(orig_config), 0)
            to_top_level =  max(BaseConfiguration.top_level(config), 0)
            self.structure.copy_data(LevelKey.TOP, 'reference',  orig_config[f'level{from_top_level}'][ControlUnitIndices.REF_INDEX], config[f'level{to_top_level}'][ControlUnitIndices.REF_INDEX])


            for level in range(from_top_level+1):
                if level < to_top_level+1:
                    level_key = self.get_level_key(level, from_top_level+1)
                    """
                    if level == to_top_level:
                        level_key = 'leveltop'
                    elif level==0 :
                        level_key = 'level0'
                    else:
                        level_key = 'leveln'
                    """
                    for item in range(min(len(orig_config[f'level{level}'][ControlUnitIndices.PER_INDEX]), len(config[f'level{level}'][ControlUnitIndices.PER_INDEX]))):
                        self.structure.copy_data(level_key, 'perception',  orig_config[f'level{level}'][ControlUnitIndices.PER_INDEX][item], config[f'level{level}'][ControlUnitIndices.PER_INDEX][item])

                    self.structure.copy_data(level_key, 'output',  orig_config[f'level{level}'][ControlUnitIndices.OUT_INDEX], config[f'level{level}'][ControlUnitIndices.OUT_INDEX])

                    for item in range(min(len(orig_config[f'level{level}'][ControlUnitIndices.REF_INDEX]),len(config[f'level{level}'][ControlUnitIndices.REF_INDEX]))):
                        if level != from_top_level and level != to_top_level:
                            #print('>',rindex,item, '<',orig_config[f'level{level}'])
                            #print('>',rindex,item, '<',config[f'level{level}'])
                            self.structure.copy_data(level_key, 'reference',  orig_config[f'level{level}'][ControlUnitIndices.REF_INDEX][item], config[f'level{level}'][ControlUnitIndices.REF_INDEX][item])

            mutant = creator.Individual(BaseConfiguration.dict_to_raw(config))

        return mutant, mutated_structure




    def create_plot(self):
        return 0

    def animate(self, epoch):
        pass



# Cell
def check_hash_file_exists(dir_path, hash_str):
    for file in os.listdir(dir_path):
        if file.find(hash_str)>=0:
            return True, file
    return False, None

# Cell
def evolve_from_properties_file(root_dir='.', path='.', file=None, verbose=None, env_name=None, seed=None, test=False,
        gens=None, pop_size=None, nevals = None, move=None, out_dir=None, draw=False, parallel=False, video_wrap=False,
        log=False, figsize=(12,12), summary=False, print_properties=False, overwrite=False, output=False):

    properties = load_properties(root_dir, path, file, print_properties=print_properties, evolve=True)

    modes_list = properties['modes']
    attr_mut_pb = properties['attr_mut_pb']

    lower_float= properties['lower_float']
    upper_float= properties['upper_float']
    levels_limit= properties['levels_limit']
    columns_limit= properties['columns_limit']
    min_levels_limit= properties['min_levels_limit']
    min_columns_limit= properties['min_columns_limit']
    references= properties['references']
    types_strings = properties['types_strings']
    configs_strings = properties['configs_strings']
    error_properties = properties['error_properties']
    if nevals == None:
        nevals = properties['nevals']

    # = properties['']
    # = properties['']

    # arch structure
    modes = {LevelKey.ZERO:modes_list[0], LevelKey.N:modes_list[1],LevelKey.TOP:modes_list[2],LevelKey.ZEROTOP :modes_list[3]}
    arch_structure = ArchitectureStructure(modes=modes)

    # structure
    sdargs={'attr_mut_pb':attr_mut_pb, 'lower_float':lower_float,
            'upper_float':upper_float, 'levels_limit': levels_limit,
            'columns_limit': columns_limit, 'min_levels_limit': min_levels_limit,
            'min_columns_limit': min_columns_limit}
    #print(sdargs)
    structure = StructureDefinition(references=references,**sdargs)
    structure.arch_structure=arch_structure

    types_string=""
    for type_key in types_strings.keys():
        types_string+=types_strings[type_key]
        type_list = stringListToListOfStrings(types_strings[type_key], ',')
        lk = eval(type_list[0])
        #print(lk, type_list[1], type_list[2])
        structure.set_config_type(lk, type_list[1], type_list[2])

    configs_string=""
    for config_key  in configs_strings.keys():
        configs_string+=configs_strings[config_key]
        config_list = stringListToListOfStrings(configs_strings[config_key],',')
        lk = eval(config_list[0])
        bk = eval(config_list[3])
        #print(lk, config_list[1], config_list[2], bk)
        structure.set_config_parameter(lk, config_list[1], config_list[2], bk)

    properties_string=""
    #print(error_properties)
    for property in error_properties:
        property_string = property[0]+property[1]
        properties_string+=property_string

    #print(structure.get_config())

    # evolve arguments
    if pop_size == None:
        pop_size = properties['pop_size']

    if gens == None:
        gens = properties['gens']

    if seed == None:
        seed = properties['seed']

    # env factory
    if env_name==None:
        env_name = properties['env_name']
    env = EnvironmentFactory.createEnvironment(env_name)
    env.set_name(env_name)
    if video_wrap:
        env.set_video_wrap(video_wrap)
        env.create_env(seed)

    if 'early_termination' in properties.keys():
        early_termination = properties['early_termination']
        env.early_termination = early_termination


    error_collector = properties['error_collector_type']
    error_collector_string=error_collector

    error_response=properties['error_response_type']
    error_response_string=error_response

    debug = 0
    if 'debug' in verbose.keys():
        debug = verbose['debug']
    hpct_verbose = False
    if 'hpct_verbose' in verbose.keys():
        hpct_verbose = verbose['hpct_verbose']

    save_arch_all = False
    if 'save_arch_all'  in verbose.keys():
        save_arch_all=verbose['save_arch_all']
    save_arch_gen = False
    if 'save_arch_gen' in verbose.keys():
        save_arch_gen=verbose['save_arch_gen']
    display_env = False
    if 'display_env' in verbose.keys():
        display_env = verbose['display_env']

    evolve_verbose=False
    if 'evolve_verbose' in  verbose.keys():
        evolve_verbose =verbose['evolve_verbose']

    deap_verbose=False
    if 'deap_verbose' in verbose.keys():
        deap_verbose=verbose['deap_verbose']

    inputs = properties['inputs']
    references = properties['references']
    top_inputs=None
    if 'top_inputs' in properties.keys():
        top_inputs=properties['top_inputs']
        structure.set_structure_parameter('top_inputs', top_inputs)

    desc = properties['desc']
    # create hash
    #print(modes_list)
    modes_string = listNumsToString(modes_list)
    hash_string =  f'{inputs}{references}{top_inputs}{properties_string}{desc}{error_response_string}{error_collector_string}{types_string}{modes_string}{seed}{pop_size}{gens}{properties["attr_mut_pb"]}{properties["structurepb"]}{properties["runs"]}{properties["lower_float"]}{properties["upper_float"]}{properties["levels_limit"]}{properties["columns_limit"]}{properties["min_levels_limit"]}{properties["min_columns_limit"]}{properties["error_limit"]}{properties["p_crossover"]}{properties["p_mutation"]}'
    #print(hash_string)
    hash_num = hashlib.md5(hash_string.encode()).hexdigest()
    #print(hash_num)

    if output:
        dir1 = out_dir + env_name
        if not os.path.isdir(dir1):
            os.mkdir(dir1)
        dir2 = dir1+os.sep+desc
        if not os.path.isdir(dir2):
            os.mkdir(dir2)
        exists, fname = check_hash_file_exists(dir2, hash_num)
        if exists and not overwrite:
            print(f'Skipping file {fname}')
            return None,None,None

    evargs={'inputs': inputs, 'env':env, 'num_actions':int(properties['num_actions']), 'nevals':nevals,
            'error_collector_type':error_collector, 'error_response_type':error_response,
            'references': references, 'structure':structure, 'structurepb':float(properties['structurepb']),
            'error_limit':float(properties['error_limit']), 'runs':int(properties['runs']), 'debug':debug,
            'seed':seed, 'hpct_verbose':hpct_verbose, 'error_properties' : error_properties}
    #print(evargs)
    stdev = DynamicEvolver(**evargs)

    evwargs={'evolver':stdev, 'pop_size':pop_size, 'p_crossover': float(properties['p_crossover']),
             'p_mutation': float(properties['p_mutation']), 'save_arch_all': save_arch_all, 'save_arch_gen':save_arch_gen,
             'display_env':display_env, 'select':{'selection_type':'tournament'} ,'parallel':parallel}
    #print(evwargs)
    evr = EvolverWrapper(**evwargs)

    if test:
        #meantime = evr.run(gens=int(db['MAX_GENERATIONS']), verbose=verbose['evolve_verbose'], deap_verbose=verbose['deap_verbose'])
        raw= [777] #evr.best()
        score = 47.2345 #evr.best_score()
        print("Best Score: %0.5f" % score)
        print("Best Ind: ", raw)
        #print(f'Mean time: {meantime:6.3f}')

    else:
        meantime, log_string = evr.run(gens=gens, verbose=evolve_verbose, deap_verbose=deap_verbose, log=log)
        raw= evr.best()
        score = evr.best_score()
        s1 = f'Best Score: {score:0.5f}'
        s2 = f'Best Ind: {raw}'
        s3 = f'Mean time: {meantime:6.3f}'
        if log:
            log_string = ''.join((log_string, '# ', s1, '\n# ', s2, '\n# ', s3, '\n'))
        if evolve_verbose:
            print(s1)
            print(s2)
            print(s3)

        # draw architecture
        if draw:
            inputs_names = None
            if 'inputs_names' in properties.keys():
                inputs_names = properties['inputs_names']
            if move == None:
                move={}
            DynamicArchitecture.draw_raw(raw, arch_structure=arch_structure, move=move, env=env, inputs_names=inputs_names,
                     inputs=inputs, top_input_indexes=top_inputs, figsize=figsize, summary=summary)

    # write results
    output_file = None
    if output:
        cs = DynamicConfigurationStructure()
        struct = cs.from_raw(raw)
        levels = len(struct['grid'])
        cols = max(struct['grid'])
        if seed == None:
            seed = 'N'
        f = open(file, "r")
        file_contents = f.read()

        output_file = dir2+os.sep +f'ga-{score:07.3f}-s{seed:03}-{levels}x{cols}-' + modes_string +f'-{hash_num}.properties'
        if print_properties:
            print(output_file)
        f = open(output_file, "w")
        dateTimeObj = datetime.now()
        f.write(f'# Date {dateTimeObj}\n')
        f.write('# Result'+'\n')
        f.write('# Best individual'+'\n')
        f.write(f'raw = {raw}'+'\n')
        f.write(f'score = {score:0.5f}'+'\n')
        f.write(f'# Time  {meantime:0.4f}'+'\n')
        f.write(file_contents.replace(f'seed = {int(properties["seed"])}', f'seed = {seed}')+'\n')
        if log:
            f.write(log_string)

        f.close()
    env.close()

    return output_file, evr, score



# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/02_configs.ipynb (unless otherwise specified).

__all__ = ['BaseConfiguration', 'ProportionalConfiguration', 'ConfigurationStructure', 'DynamicConfigurationStructure',
           'DynamicConfiguration']

# Cell
import gym
import numpy as np
import random
from abc import ABC, abstractmethod

from pct.environments import PendulumV0
from pct.environments import VelocityModel
from pct.functions import IndexedParameter
from pct.functions import Constant
from pct.architectures import ProportionalArchitecture
from pct.architectures import DynamicArchitecture
from .structure import StructureDefinition
from pct.errors import BaseErrorCollector
from pct.architectures import LevelKey
from pct.environments import DummyModel
from epct.structure  import BinaryOnes
from pct.structure import ArchitectureStructure

# Cell
class BaseConfiguration(ABC):
    "Base class of an array configuration. This class is not used direclty by developers, but defines the functionality common to all."
    def __init__(self, name, num_inputs, num_actions, grid, references):
        self.name = name
        self.num_inputs=num_inputs
        self.num_actions=num_actions
        self.grid=grid
        self.references=references
        self.config={}
        self.config['parameters']={}
        if grid[-1] != len(references):
            raise ValueError(f'Number of references {len(references)} must equal number of top-level columns {grid[-1]}')

    def __call__(self):
        level0 = self.level0config()
        if len(self.grid) == 1:
            level0.append(self.references)
        self.config['level0']=level0

        if len(self.grid) > 2:
            for level in range(1,len(self.grid), 1):
                numColumnsThisLevel=self.grid[level]
                numColumnsPreviousLevel=self.grid[level-1]
                self.config[f'level{level}']=self.levelnconfig(numColumnsThisLevel, numColumnsPreviousLevel)

        if len(self.grid) > 1:
            self.config[f'level{len(self.grid)-1}'] = self.leveltopconfig(self.references, self.grid[-2])

        return self.config

    @classmethod
    def top_level(cls, config):
        levels = len(config.keys())
        if 'parameters' in config.keys():
            levels -= 1

        return levels-1

    def lists_of_ones(self, num_lists, length):
        wts=[]
        for i in range(num_lists):
            wt = [1 for iter in range(length)]

            wts.append(wt)

        return wts

    def create_binary_lists(self, num_lists, length):

        wts=[]
        for i in range(num_lists):
            if length==1 and num_lists==1:
                wt = [1]
            else:
                wt = [random.randint(0, 1) for iter in range(length)]
                if np.sum(wt) == 0:
                    index = random.randint(0, length-1)
                    wt[index] = 1

            wts.append(wt)


        #if length>1 and num_lists==1 and np.sum(wts) == 0:
        #    index = random.randint(0, length-1)
        #    wts[0][index] = 1

        # flip one list if all 0
        if length==1 and num_lists>1 and np.sum(wts) == 0:
            index = random.randint(0, num_lists-1)
            wts[index] = [1]

        return wts

    def create_randomint_lists(self, lower, upper, num_lists, length):
        wts=[]
        for i in range(num_lists):
            wt = [random.randint(lower, upper) for iter in range(length)]
            wts.append(wt)
        return wts

    def create_randomfloat_list(self, lower, upper, length):
        return [random.uniform(lower, upper) for iter in range(length)]

    def create_randomfloat_lists(self, lower, upper, num_lists, length):
        wts=[]
        for i in range(num_lists):
            wt = [random.uniform(lower, upper) for iter in range(length)]
            wts.append(wt)
        return wts

    def to_raw(self):
        raw = []
        for key in self.config.keys():
            if key != 'parameters':
                raw.append(self.config[key])
        return raw

    @classmethod
    def from_raw(cls, raw):
        config = {}
        config['parameters']={}
        for level in range(len(raw)):
            config[f'level{level}'] = raw[level]
        return config

    @classmethod
    def dict_to_raw(cls, config):
        raw = []
        for key in config.keys():
            if key != 'parameters':
                raw.append(config[key])
        return raw


    @abstractmethod
    def level0config(self):
         pass

    @abstractmethod
    def levelnconfig(self, numColumnsThisLevel, numColumnsPreviousLevel):
         pass

    @abstractmethod
    def leveltopconfig(self, top_references, numColumnsPreviousLevel):
         pass

    @classmethod
    def copy_data(cls, from_config, to_config):
        levels = len(from_config)-1
        for level in range(levels):
            index=0
            for item in range(len(from_config[f'level{level}'][index])):
                if item < len(to_config[f'level{level}'][index]):
                    for element in range(len(from_config[f'level{level}'][index][item])):
                        if element < len(to_config[f'level{level}'][index][item]):
                            to_config[f'level{level}'][index][item][element]= from_config[f'level{level}'][index][item][element]
                    BaseConfiguration.ensure_non_zero(to_config, level, index, item)

            index=1
            for item in range(len(from_config[f'level{level}'][index])):
                if item < len(to_config[f'level{level}'][index]):
                    to_config[f'level{level}'][index][item]= from_config[f'level{level}'][index][item]

            index=2
            for item in range(len(from_config[f'level{level}'][index])):
                #print(level, item)
                if item < len(to_config[f'level{level}'][index]):
                    for element in range(len(from_config[f'level{level}'][index][item])):
                        #print(level, item, element)
                        if element < len(to_config[f'level{level}'][index][item]):
                            #print(to_config[f'level{level}'][index][item][element])
                            #print(from_config[f'level{level}'][index][item][element])
                            to_config[f'level{level}'][index][item][element]= from_config[f'level{level}'][index][item][element]
                    BaseConfiguration.ensure_non_zero(to_config, level, index, item)


    @classmethod
    def copy_data_from_top(cls, from_config, to_config):
        levels = len(from_config)-1
        to_level = levels
        skip_level=levels-1

        for level in range(levels-1, -1, -1):
            if to_level==skip_level:
                to_level-=1
            #print(level, to_level, skip_level)
            index=0
            for item in range(len(from_config[f'level{level}'][index])):
                for element in range(len(from_config[f'level{level}'][index][item])):
                    to_config[f'level{to_level}'][index][item][element]= from_config[f'level{level}'][index][item][element]

            index=1
            for item in range(len(from_config[f'level{level}'][index])):
                to_config[f'level{to_level}'][index][item]= from_config[f'level{level}'][index][item]

            index=2
            for item in range(len(from_config[f'level{level}'][index])):
                for element in range(len(from_config[f'level{level}'][index][item])):
                    to_config[f'level{to_level}'][index][item][element]= from_config[f'level{level}'][index][item][element]
            to_level-=1



    @classmethod
    def copy_data_from_larger(cls, from_config, to_config):
        levels = len(from_config)-1
        to_level = levels-2
        skip_level=levels-2
        #print('skip_level',skip_level)
        for level in range(levels-1, -1, -1):
            if level==skip_level:
                pass #level-=1
            else:
                #print(level, to_level, skip_level)
                index=0
                for item in range(len(to_config[f'level{to_level}'][index])):
                    if to_level ==0 and len(to_config[f'level{to_level}'][index][item]) > 1:
                        for element in range(len(to_config[f'level{to_level}'][index][item])):
                            to_config[f'level{to_level}'][index][item][element]= from_config[f'level{level}'][index][item][element]

                index=1
                for item in range(len(to_config[f'level{to_level}'][index])):
                    to_config[f'level{to_level}'][index][item]= from_config[f'level{level}'][index][item]

                index=2
                for item in range(len(from_config[f'level{level}'][index])):
                    if item < len(to_config[f'level{to_level}'][index]):
                        for element in range(len(to_config[f'level{to_level}'][index][item])):
                            to_config[f'level{to_level}'][index][item][element]= from_config[f'level{level}'][index][item][element]

                to_level-=1
            if to_level < 0:
                break

    @classmethod
    def ensure_non_zero(cls, config, level, index, item):
        if np.sum(config[f'level{level}'][index][item])==0:
            length = len(config[f'level{level}'][index][item])
            ind = random.randint(0, length-1)
            config[f'level{level}'][index][item][ind]=1


    @classmethod
    def copy_data_from_smaller(cls, from_config, to_config):
        levels = len(from_config)-1
        to_level = 0
        skip_level=levels-1
        #print(to_config)
        for level in range(levels):
            if to_level==skip_level:
                to_level+=1
            #print(level, to_level, skip_level)
            index=0
            #print(len(from_config[f'level{level}'][index]), len(to_config[f'level{to_level}'][index]) )

            for item in range(len(from_config[f'level{level}'][index])):
                #print('item', item)
                #if item < len(to_config[f'level{to_level}'][index]) and item < len(from_config[f'level{level}'][index]):
                if item < len(to_config[f'level{to_level}'][index]) :
                    #print(len(from_config[f'level{level}'][index][item]), len(to_config[f'level{to_level}'][index][item]) )

                    for element in range(len(from_config[f'level{level}'][index][item])):
                        #print('elem', element)
                        if element < len(to_config[f'level{to_level}'][index][item]) and element < len(from_config[f'level{level}'][index][item]) :
                            to_config[f'level{to_level}'][index][item][element]= from_config[f'level{level}'][index][item][element]
                    BaseConfiguration.ensure_non_zero(to_config, to_level, index, item)
                #if np.sum(to_config[f'level{to_level}'][index][item])==0:
                #    length = len(to_config[f'level{to_level}'][index][item])
                #    ind = random.randint(0, length-1)
                #    to_config[f'level{to_level}'][index][item][ind]=1

            index=1
            for item in range(len(from_config[f'level{level}'][index])):
                to_config[f'level{to_level}'][index][item]= from_config[f'level{level}'][index][item]

            index=2
            for item in range(len(from_config[f'level{level}'][index])):
                if item < len(to_config[f'level{to_level}'][index]):
                    for element in range(len(from_config[f'level{level}'][index][item])):
                        to_config[f'level{to_level}'][index][item][element]= from_config[f'level{level}'][index][item][element]
            to_level+=1


# Cell
class ProportionalConfiguration(BaseConfiguration):
    "ProportionalConfiguration"
    def __init__(self, name="proportional", num_inputs=1, num_actions=1, grid=[2, 2], references=[1],
                 lower_float=-100, upper_float=100, **cargs):
        super().__init__(name, num_inputs, num_actions, grid, references)
        self.lower_float=lower_float
        self.upper_float=upper_float
        self.config['parameters']['lower_float']=lower_float
        self.config['parameters']['upper_float']=upper_float


    def level0config(self):
        numColumnsThisLevel = self.grid[0]
        # inputs
        iwts=self.create_binary_lists(self.num_inputs, numColumnsThisLevel)

        # outputs
        owts=self.create_randomfloat_list(self.lower_float, self.upper_float, numColumnsThisLevel)

        # actions
        #actwts = self.lists_of_ones(self.num_actions, numColumnsThisLevel)
        actwts = self.create_binary_lists(self.num_actions, numColumnsThisLevel)

        lwts=[]
        lwts.append(iwts)
        lwts.append(owts)
        lwts.append(actwts)

        return lwts

    def levelnconfig(self, numColumnsThisLevel, numColumnsPreviousLevel):
        # inputs
        iwts=self.create_binary_lists(numColumnsThisLevel, numColumnsPreviousLevel)

        # outputs
        owts=self.create_randomfloat_list(self.lower_float, self.upper_float, numColumnsThisLevel)

        # lower refs
        rwts = self.create_randomfloat_lists(self.lower_float, self.upper_float, numColumnsPreviousLevel, numColumnsThisLevel)

        lwts=[]
        lwts.append(iwts)
        lwts.append(owts)
        lwts.append(rwts)

        return lwts


    def leveltopconfig(self, top_references, numColumnsPreviousLevel):

        numColumnsThisLevel=len(top_references)
        # inputs
        iwts=self.create_binary_lists(numColumnsThisLevel, numColumnsPreviousLevel)

        # outputs
        owts=self.create_randomfloat_list(self.lower_float, self.upper_float, numColumnsThisLevel)

        # lower refs
        rwts = self.create_randomfloat_lists(self.lower_float, self.upper_float, numColumnsPreviousLevel, numColumnsThisLevel)

        lwts=[]
        lwts.append(iwts)
        lwts.append(owts)
        lwts.append(rwts)
        lwts.append(top_references)

        return lwts



# Cell
class ConfigurationStructure():

    def __init__(self, structure=None, **cargs):
        if structure==None:
            self.structure={}
        else:
            self.structure=structure

    def from_raw(self, raw):

        config = BaseConfiguration.from_raw(raw)
        levels = len(config)-1
        num_inputs=len(config['level0'][0])
        self.structure['num_inputs']=num_inputs
        num_actions=len(config['level0'][2])
        self.structure['num_actions']=num_actions

        grid=[]
        for level in range(levels):
            grid.append(len(config[f'level{level}'][1]))
        self.structure['grid']=grid
        references = config[f'level{levels-1}'][3]
        self.structure['references']=references
        return self.structure

    def add_nodes(self, level, num_nodes=1):
        mutated_structure = 0

        levels = len(self.structure['grid'])
        if level < levels-1:
            self.structure['grid'][level]+=num_nodes
            mutated_structure = 1

        return self.structure, mutated_structure

    def remove_nodes(self, level, num_nodes=1):
        mutated_structure = 0

        levels = len(self.structure['grid'])
        if level == levels-1:
            raise Exception("Cannot remove nodes from the top level.")

        if level > levels-1:
            raise Exception(f'Level {level} does not exist for hierarchy with levels {levels}.')

        existing_num_nodes = self.structure['grid'][level]
        if num_nodes > existing_num_nodes-1:
            raise Exception(f'Cannot remove {num_nodes} nodes from level with {existing_num_nodes} nodes.')

        self.structure['grid'][level]-=num_nodes
        mutated_structure = 1

        return self.structure, mutated_structure

    def add_level(self, num_columns=1):
        mutated_structure = 1

        #levels = len(self.structure['grid'])
        self.structure['grid'].insert(-1, num_columns)

        return self.structure, mutated_structure

    def remove_level(self):
        mutated_structure = 0

        levels = len(self.structure['grid'])
        if levels>1:
            del self.structure['grid'][levels-1:]
            self.structure['grid'][-1]=len(self.structure['references'])
            mutated_structure = 1

        return self.structure, mutated_structure


# Cell

class DynamicConfigurationStructure(ConfigurationStructure):

    def __init__(self, structure=None, **cargs):
        if structure==None:
            self.structure={}
        else:
            self.structure=structure

    def from_raw(self, raw):

        config = BaseConfiguration.from_raw(raw)
        levels = len(config)-1
        num_inputs=len(config['level0'][0])
        self.structure['num_inputs']=num_inputs
        num_actions=len(config['level0'][3])
        self.structure['num_actions']=num_actions

        grid=[]
        for level in range(levels):
            grid.append(len(config[f'level{level}'][1]))
        self.structure['grid']=grid
        references = config[f'level{levels-1}'][2]
        self.structure['references']=references
        return self.structure



# Cell
class DynamicConfiguration(BaseConfiguration):
    "DynamicConfiguration"
    def __init__(self, inputs=None, num_inputs=1, num_actions=1, references=None,
                 structure=None, grid=None,  **cargs):

        #self.inputs=inputs
        num_top_inputs=0

        top_inputs = structure.get_parameter('top_inputs')
        if top_inputs != None:
            num_top_inputs = len(top_inputs)
            if inputs != None:
                num_inputs = len(inputs)

        self.references=references
        if structure == None:
            structure = StructureDefinition(references=references)

        self.num_inputs=num_inputs - num_top_inputs
        self.num_top_inputs=num_top_inputs
        self.num_actions=num_actions
        self.config={}
        self.structure=structure
        self.grid=grid
        """
        self.lower_float=lower_float
        self.upper_float=upper_float
        self.config['parameters']['lower_float']=lower_float
        self.config['parameters']['upper_float']=upper_float
        """

    def __call__(self):

        if self.grid==None:
            grid=[]
            levels = random.randint(self.structure.get_parameter('min_levels_limit'), self.structure.get_parameter('levels_limit'))
            for level in range(levels-1):
                columns = random.randint(self.structure.get_parameter('min_columns_limit'), self.structure.get_parameter('columns_limit'))
                grid.append(columns)
            grid.append(len(self.references))
        else:
            grid=self.grid


        levels = len(grid)
        #print(grid)
        if levels == 1:
            level0 = self.level0topconfig(grid[0])
            self.config['level0']=level0
            return self.config
        else:
            level0 = self.level0config(grid[0], grid[1])

        self.config['level0']=level0

        if levels > 2:
            for level in range(1,levels-1, 1):
                numColumnsThisLevel=grid[level]
                numColumnsPreviousLevel=grid[level-1]
                numColumnsNextLevel = grid[level+1]
                self.config[f'level{level}']=self.levelnconfig(numColumnsThisLevel, numColumnsNextLevel, numColumnsPreviousLevel)

        if levels > 1:
            self.config[f'level{len(grid)-1}'] = self.leveltopconfig(grid[-1], grid[-2])

        return self.config

    def level0config(self, numColumnsThisLevel, numColumnsNextLevel):
        return self.structure.get_level0(self.num_inputs,numColumnsThisLevel, numColumnsNextLevel, self.num_actions)

    def level0topconfig(self, num_columns):
        return self.structure.get_level0top(self.num_inputs+self.num_top_inputs,num_columns, self.num_actions)

    def levelnconfig(self, numColumnsThisLevel, numColumnsNextLevel, numColumnsPreviousLevel):
        return self.structure.get_leveln(numColumnsThisLevel, numColumnsNextLevel, numColumnsPreviousLevel)

    def leveltopconfig(self, numColumnsThisLevel, numColumnsPreviousLevel):
        return self.structure.get_leveltop(numColumnsThisLevel, numColumnsPreviousLevel, self.num_top_inputs)

    @classmethod
    def convert_to_raw_from_proportional_raw(cls, prop_raw, verbose=False):
        if verbose:
            #print(prop_raw)
            for level in range(len(prop_raw)):
                if level == 0:
                    print(level, 'P', prop_raw[level][0],'O', prop_raw[level][1],'A', prop_raw[level][2])
                elif level == len(prop_raw)-1:
                    print(level, 'P', prop_raw[level][0],'O', prop_raw[level][1],'R', prop_raw[level][2],'MR', prop_raw[level][3])
                else:
                    print(level, 'P', prop_raw[level][0],'O', prop_raw[level][1],'R', prop_raw[level][2])


        levels = len(prop_raw)
        if levels == 0:
            dyn_raw=prop_raw
        else:
            dyn_raw=[]
            for level in range(len(prop_raw)):
                level_raw=[]
                if level == 0:
                    level_raw.append(prop_raw[level][0])
                else:
                    plist=[]
                    for item in range(len(prop_raw[level][0][0])):
                        item_list = []
                        for list in prop_raw[level][0]:
                            item_list.append(list[item])
                        plist.append(item_list)
                    level_raw.append(plist)
                level_raw.append(prop_raw[level][1])
                if level == 0:
                    level_raw.append(prop_raw[level+1][2])
                if level == levels-1:
                    level_raw.append(prop_raw[level][3])
                else:
                    level_raw.append(prop_raw[level][2])
                dyn_raw.append(level_raw)

        if verbose:
            print()

            for level in range(len(dyn_raw)):
                if level == 0:
                    print(level, 'P', dyn_raw[level][0],'O', dyn_raw[level][1],'R', dyn_raw[level][2],'A', dyn_raw[level][3])
                else:
                    print(level, 'P', dyn_raw[level][0],'O', dyn_raw[level][1],'R', dyn_raw[level][2])


        return dyn_raw




# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/03_structure.ipynb (unless otherwise specified).

__all__ = ['BinaryOnes', 'BaseParameterType', 'Binary', 'Float', 'Literal', 'Smooth', 'ParameterFactory',
           'StructureDefinition']

# Cell
import random
import enum
import numpy as np
from abc import ABC
from deap import tools

from pct.functions import WeightedSum
from pct.architectures import LevelKey
from pct.structure import ArchitectureStructure

# Cell
class BinaryOnes(enum.Enum):
   ALLOW_ALL_ZEROS = 0
   ALL_ONES = 1
   AT_LEAST_ONE = 2

# Cell
class BaseParameterType(ABC):
    "Base class of a hierarchy parameter type. This class is not used direclty by developers, but defines the interface common to all."
    def __init__(self):
        pass

    def set_parameters(self, pars, globals):
        for par in pars:
            cmd = f'self.{par} = {pars[par]}'
            exec(cmd)

    def set_node_function(self, node, function, thislevel, targetlevel, targetprefix, column, num_target_indices, inputs, input_weights, by_column):
        func = node.get_function(function)
        prefix = function[0].capitalize()
        func.set_name(f'{prefix}L{thislevel}C{column}ws')

        """
        print('Base',func.get_name())
        print('Base',inputs)
        print('Base',input_weights)
        print('Base',column)
        print('Base',num_target_indices)
        """
        weights=[]
        for inputIndex in range(num_target_indices):
            if inputs==None:
                name=f'{targetprefix}L{targetlevel}C{inputIndex}ws'
            else:
                name=inputs[inputIndex]
            func.add_link(name)
            #print(name)
            if by_column:
                weights.append(input_weights[column][inputIndex])
            else:
                #print(inputIndex,column)
                weights.append(input_weights[inputIndex][column])
        func.weights=np.array(weights)

    def copy_data(self, from_wts, to_wts):

        #print(from_wts, to_wts)
        from_len=len(from_wts)
        to_len = len(to_wts)

        for ctr in range(min(from_len, to_len)):
            to_wts[ctr] = from_wts[ctr]

    def has_changed(self, wts1, wts2):
        mutated = 0
        if wts1 != wts2:
            mutated = 1
        return mutated



class Binary(BaseParameterType):

    def __init__(self):
        self.ones=BinaryOnes.AT_LEAST_ONE
        self.attr_mut_pb=None
        self.attr_cx_uniform_pb=None

    def get_weights_list(self, num_inputs, num_columns):
        wts_list=[]
        for i in range(num_inputs):
            if self.ones == BinaryOnes.ALL_ONES:
                wts = [1] * num_columns
            else:
                wts = [random.randint(0, 1) for iter in range(num_columns)]
                if self.ones == BinaryOnes.AT_LEAST_ONE:
                    if np.sum(wts) == 0:
                        index = random.randint(0, num_columns-1)
                        wts[index] = 1

            wts_list.append(wts)

        return wts_list


    def get_weights_list_assymetric(self, num_top_inputs, numColumnsPreviousLevel, numColumnsThisLevel):
        wts_list=[]
        num_columns=1
        for i in range(num_top_inputs):
            top_wts = [1] * num_columns
            wts_list.append(top_wts)

        num_columns = numColumnsThisLevel - num_top_inputs
        if num_columns > 0:
            for i in range(numColumnsPreviousLevel- num_top_inputs):
                if self.ones == BinaryOnes.ALL_ONES:
                    wts = [1] * num_columns
                else:
                    wts = [random.randint(0, 1) for iter in range(num_columns)]
                    if self.ones == BinaryOnes.AT_LEAST_ONE:
                        if np.sum(wts) == 0:
                            index = random.randint(0, num_columns-1)
                            wts[index] = 1

                wts_list.append(wts)

        #print('assymetric',wts_list)
        return wts_list


    def set_parameters(self, pars, globals):
        super().set_parameters(pars, globals)
        if self.attr_mut_pb==None:
            self.attr_mut_pb= globals['attr_mut_pb']
        if self.attr_cx_uniform_pb==None:
            self.attr_cx_uniform_pb= globals['attr_cx_uniform_pb']


    def mutate(self, wts):
        if self.ones == BinaryOnes.ALL_ONES or (self.ones == BinaryOnes.AT_LEAST_ONE and len(wts)==1):
            pass
        else:
            weights, = tools.mutFlipBit(wts, self.attr_mut_pb)
            if self.ones == BinaryOnes.AT_LEAST_ONE:
                if np.sum(weights) == 0:
                    index = random.randint(0, len(weights)-1)
                    weights[index] = 1

    def mutate_binary_lists(self, lists):
        mut_pb = 0.5
        list = []
        #print('***1',lists)
        for item in lists:
            list.append(item[0])
        #print('***2',list)
        list, = tools.mutFlipBit(list, mut_pb)
        #print('***3',list)
        if np.sum(list) == 0:
            index = random.randint(0, len(list)-1)
            list[index] = 1
        #print('***4',list)

        for i in range(len(list)):
            lists[i][0]=list[i]

        #print('***5',lists)


    def ensure_non_zero(self, lists):
        if np.sum(lists) == 0:
            #print('ensure_non_zero 1 ', lists)
            n_lists = len(lists)
            list_len = len(lists[0])
            total_items = n_lists * list_len
            index = random.randint(0, total_items-1)
            #print(index)
            index1 = (int)(index/list_len)
            index2 = index - index1*list_len
            #print(index1)
            #print(index2)
            lists[index1][index2] = 1
            #print('ensure_non_zero 2 ', lists)


    def mate(self, wts1, wts2):
        #print('Binary mate')
        #print(wts2)
        #print(wts2)
        wts1, wts2 = tools.cxUniform(wts1, wts2, self.attr_cx_uniform_pb)


    def copy_data(self, from_wts, to_wts):

        #print(from_wts, to_wts)
        from_len=len(from_wts)
        to_len = len(to_wts)

        for ctr in range(min(from_len, to_len)):
            to_wts[ctr] = from_wts[ctr]

        if self.ones == BinaryOnes.AT_LEAST_ONE:
            if np.sum(to_wts) == 0:
                index = random.randint(0, len(to_wts)-1)
                to_wts[index] = 1


        #return to_wts


    class Factory:
        def create(self): return Binary()


class Float(BaseParameterType):

    def __init__(self):
        self.lower=None
        self.upper=None
        self.mu=None
        self.sigma=None
        self.attr_mut_pb=None
        self.alpha=None

    def get_weights_list(self,  num_lists, length):
        wts=[]
        for i in range(num_lists):
            wt = [random.uniform(self.lower, self.upper) for iter in range(length)]
            wts.append(wt)
        return wts

    def get_weights_list_assymetric(self, num_top_inputs, numColumnsPreviousLevel, numColumnsThisLevel):
        wts_list=[]
        num_columns=1
        for i in range(num_top_inputs):
            top_wts = [1] * num_columns
            wts_list.append(top_wts)

        num_columns = numColumnsThisLevel - num_top_inputs
        if num_columns > 0:
            for i in range(numColumnsPreviousLevel- num_top_inputs):
                wts = [random.uniform(self.lower, self.upper) for iter in range(num_columns)]
                wts_list.append(wts)

        #print('assymetric',wts_list)
        return wts_list


    def set_parameters(self, pars, globals):
        super().set_parameters(pars, globals)
        if self.lower==None:
            self.lower = globals['lower_float']
        if self.upper==None:
            self.upper = globals['upper_float']
        if self.mu==None:
            self.mu = globals['mu']
        if self.sigma==None:
            self.sigma= globals['sigma']
        if self.alpha==None:
            self.alpha= globals['alpha']
        if self.attr_mut_pb==None:
            self.attr_mut_pb= globals['attr_mut_pb']

    def mutate(self, wts):
        weights, = tools.mutGaussian(wts, mu=self.mu, sigma=self.sigma, indpb=self.attr_mut_pb)

    def mate(self, wts1, wts2):
        #print('Float mate')
        #print(wts1, wts2)
        wts1, wts2 = tools.cxBlend(wts1, wts2, self.alpha)
        #print(wts1, wts2)


    class Factory:
        def create(self): return Float()


class Literal(BaseParameterType):

    def __init__(self):
        self.value=None

    def get_weights_list(self,  num_lists, length):
        return self.value

    def set_node_function(self, node, function, thislevel, targetlevel, not_used, column, not_used1, inputs, weights, not_used2):

        func = node.get_function(function)
        prefix = function[0].capitalize()
        func.set_name(f'{prefix}L{thislevel}C{column}c')
        func.set_value(weights[column])
        """
        print('Literal',inputs)
        print('Literal',weights)
        prefix = function[0].capitalize()
        constant = Constant(weights[column], name=f'{prefix}L{thislevel}C{column}c')
        node.replace_function(function, constant, 0)
        """

    def copy_data(self, from_wts, to_wts):
        pass

    class Factory:
        def create(self): return Literal()

class Smooth(BaseParameterType):

    def __init__(self):
        self.lower=None
        self.upper=None
        self.mu=None
        self.sigma=None
        self.attr_mut_pb=None
        self.alpha=None

    def get_weights_list(self,  num_lists, length):
        wts=[]
        for i in range(num_lists):
            wt = [[random.uniform(self.lower, self.upper), random.uniform(0, 1)] for iter in range(length)]
            wts.append(wt)
        return wts

    def set_parameters(self, pars, globals):
        super().set_parameters(pars, globals)
        if self.lower==None:
            self.lower = globals['lower_float']
        if self.upper==None:
            self.upper = globals['upper_float']
        if self.mu==None:
            self.mu = globals['mu']
        if self.sigma==None:
            self.sigma= globals['sigma']
        if self.alpha==None:
            self.alpha= globals['alpha']
        if self.attr_mut_pb==None:
            self.attr_mut_pb= globals['attr_mut_pb']

    def mutate(self, wts):
        for i in range(len(wts)):
            [wts[i][0]], = tools.mutGaussian([wts[i][0]], mu=self.mu, sigma=self.sigma, indpb=self.attr_mut_pb)

        for i in range(len(wts)):
            [wts[i][1]], = tools.mutGaussian([wts[i][1]], mu=0.1, sigma=0.1, indpb=self.attr_mut_pb)
            if wts[i][1] > 1:
                wts[i][1] = 1 - (wts[i][1] - 1)

            if wts[i][1] < 1:
                wts[i][1] = abs(wts[i][1])

            if wts[i][1] > 1 or wts[i][1] < 0:
                print(wts)
                raise Exception(f'The smooth factor should be between 0 and 1. For wts it is {wts[i][1]}.')



    def mate(self, wts1, wts2):
        loops = min(len(wts1), len(wts2))
        for i in range(loops):
            [wts1[i][0]], [wts2[i][0]] = tools.cxBlend([wts1[i][0]], [wts2[i][0]], self.alpha)

        for i in range(loops):
            [wts1[i][1]], [wts2[i][1]] = tools.cxBlend([wts1[i][1]], [wts2[i][1]], 0)
            if wts1[i][1] > 1 or wts1[i][1] < 0:
                print(wts1, wts2)
                raise Exception(f'The smooth factor should be between 0 and 1. For wts1 it is {wts1[i][1]}.')
            if wts2[i][1] > 1 or wts2[i][1] < 0:
                print(wts1, wts2)
                raise Exception(f'The smooth factor should be between 0 and 1. For wts2 it is {wts2[i][1]}.')


    """
    def mutate(self, wts):
        weights, = tools.mutGaussian(wts, mu=self.mu, sigma=self.sigma, indpb=self.attr_mut_pb)

    def mate(self, wts1, wts2):
        #print('Float mate')
        #print(wts1, wts2)
        wts1, wts2 = tools.cxBlend(wts1, wts2, self.alpha)
        #print(wts1, wts2)

    def copy_data(self, from_wts, to_wts):

        #print(from_wts, to_wts)
        from_len=len(from_wts)
        to_len = len(to_wts)

        for ctr in range(min(from_len, to_len)):
            to_wts[ctr] = from_wts[ctr]
    """

    class Factory:
        def create(self): return Smooth()

class ParameterFactory:
    factories = {}
    def addFactory(id, parameterFactory):
        ParameterFactory.factories.put[id] = parameterFactory
    addFactory = staticmethod(addFactory)

    # A Template Method:
    def createParameter(id):
        if not ParameterFactory.factories.__contains__(id):
            ParameterFactory.factories[id] = \
              eval(id + '.Factory()')
        return ParameterFactory.factories[id].create()

    createParameter = staticmethod(createParameter)

# Cell
class StructureDefinition():
    "StructureDefinition"
    def __init__(self, references=None, config=None, attr_mut_pb=None, lower_float=None, upper_float=None,
                 levels_limit=None, columns_limit=None, min_levels_limit=None, min_columns_limit=None,
                 sigma=None, mu=None, alpha=None, arch_structure=None, **cargs):
        if config==None:
            self.config={'parameters': {'lower_float': -1, 'upper_float': 1,
                         #'modes' : {LevelKey.ZERO:3, LevelKey.N:3,LevelKey.TOP:4,LevelKey.ZEROTOP :4},
                         'levels_limit': 3, 'columns_limit': 3, 'min_levels_limit': 1, 'min_columns_limit': 1,
                         'attr_mut_pb': 1.0, 'sigma': 0.8, 'mu': 0.5, 'alpha':0.6, 'attr_cx_uniform_pb':0.5},
                         LevelKey.ZERO: {'perception': {'type': 'Binary'}, 'output': {'type': 'Float'}, 'reference': {'type': 'Float'}, 'action': {'type': 'Binary'}},
                         LevelKey.N: {'perception': {'type': 'Binary'}, 'output': {'type': 'Float'}, 'reference': {'type': 'Float'}},
                         LevelKey.TOP: {'perception': {'type': 'Binary'}, 'output': {'type': 'Float'}, 'reference': {'type': 'Literal'}}}
        else:
            self.config=config

        if arch_structure==None:
            self.arch_structure = ArchitectureStructure()

        if references!=None:
            self.set_config_parameter(LevelKey.TOP , 'reference', 'value', references )
        if attr_mut_pb != None:
            self.config['parameters']['attr_mut_pb']=attr_mut_pb
        if lower_float!=None:
            self.config['parameters']['lower_float']=lower_float
        if upper_float!=None:
            self.config['parameters']['upper_float']=upper_float
        if levels_limit!=None:
            self.config['parameters']['levels_limit']=levels_limit
        if columns_limit!=None:
            self.config['parameters']['columns_limit']=columns_limit
        if min_levels_limit!=None:
            self.config['parameters']['min_levels_limit']=min_levels_limit
        if min_columns_limit!=None:
            self.config['parameters']['min_columns_limit']=min_columns_limit
        if sigma!=None:
            self.config['parameters']['sigma']=sigma
        if mu!=None:
            self.config['parameters']['mu']=mu
        if alpha!=None:
            self.config['parameters']['alpha']=alpha
        #if modes!=None:
        #    self.config['parameters']['modes']=modes



    def get_arch_structure(self):
        return self.arch_structure

    def get_config(self):
        return self.config

    def set_config_type(self, level_key=None, function=None, type=None):
        ttype={'type': type}
        self.config[level_key][function]=ttype

    def set_structure_parameter(self, key=None, value=None):
        self.config['parameters'][key]=value

    def set_level_parameter(self, level=None, function=None, key=None, value=None):
        self.config[level][function][key]=value


    def set_config_parameter(self, level_key=LevelKey.N, function=None,  parameter_type=None, parameter_value=None):
        if not 'pars' in self.config[level_key][function]:
            self.config[level_key][function]['pars']={} #pars={'pars'}
        self.config[level_key][function]['pars'][parameter_type]=parameter_value

    def get_config_parameter(self, level_key=LevelKey.N, function=None,  parameter_type=None):
        if not 'pars' in self.config[level_key][function]:
            return False
        return self.config[level_key][function]['pars'][parameter_type]

    def get_level0_config(self):
        return self.config['level0']

    def get_leveln_config(self):
        return self.config['leveln']

    def get_leveltop_config(self):
        return self.config['leveltop']

    def get_type(self, level_key, function):
        package = self.config[level_key][function]
        pars={}
        if 'pars' in package.keys():
            pars = package['pars']

        return package['type'], pars

    def get_type_parameters(self, level, function):
        return self.config[level][function]

    def get_parameter(self, key):
        rtn = None
        if key in self.config['parameters'].keys():
            rtn = self.config['parameters'][key]

        return rtn

    """
    def set_node_function(self, node, function, levelkey, thislevel, targetlevel, targetprefix, column, num_target_indices, inputs, input_weights, by_column):
        type, type_parameters = self.get_type(levelkey, function)
        parameter = ParameterFactory.createParameter(type)
        parameter.set_node_function(node, function,  thislevel, targetlevel, targetprefix, column, num_target_indices, inputs, input_weights, by_column)
    """


    def get_list(self, level_key, function,  num_lists, num_items):
        parameter = self.get_parameter_object(level_key, function)
        return parameter.get_weights_list(num_lists, num_items)

    def get_list_assymetric(self, level_key, function,  num_top_inputs, numColumnsPreviousLevel, numColumnsThisLevel):
        parameter = self.get_parameter_object(level_key, function)
        return parameter.get_weights_list_assymetric(num_top_inputs, numColumnsPreviousLevel, numColumnsThisLevel)

    def mutate_list(self, level, function, wts):
        parameter = self.get_parameter_object(level, function)
        return parameter.mutate(wts)

    def mutate_binary_lists(self, level, function, lists):
        parameter = self.get_parameter_object(level, function)
        parameter.mutate_binary_lists(lists)

    def mate_lists(self, level, function, wts1, wts2):
        parameter = self.get_parameter_object(level, function)
        return parameter.mate(wts1, wts2)

    def ensure_non_zero(self, level, function, wts):
        parameter = self.get_parameter_object(level, function)
        return parameter.ensure_non_zero(wts)

    def copy_data(self, level, function, from_wts, to_wts):
        parameter = self.get_parameter_object(level, function)
        parameter.copy_data(from_wts, to_wts)

    def get_parameter_object(self, level_key, function):
        type, type_parameters = self.get_type(level_key, function)

        parameter = ParameterFactory.createParameter(type)
        parameter.set_parameters(type_parameters, self.config['parameters'])

        return parameter


    def get_level0(self, num_inputs, numColumnsThisLevel, numColumnsNextLevel, num_actions):
        config0=[]

        perception_list = self.get_list(LevelKey.ZERO,'perception', num_inputs, numColumnsThisLevel)
        output_list = self.get_list(LevelKey.ZERO,'output', num_actions, numColumnsThisLevel)
        reference_list = self.get_list(LevelKey.ZERO,'reference', numColumnsThisLevel, numColumnsNextLevel)
        action_list = self.get_list(LevelKey.ZERO,'action', num_actions, numColumnsThisLevel)

        config0.append(perception_list)
        config0.append(output_list[0])
        config0.append(reference_list)
        config0.append(action_list)

        return config0


    def get_leveln(self, numColumnsThisLevel, numColumnsNextLevel, numColumnsPreviousLevel):
        config=[]

        perception_list = self.get_list(LevelKey.N,'perception', numColumnsPreviousLevel, numColumnsThisLevel)

        output_list = self.get_list(LevelKey.N,'output', numColumnsPreviousLevel, numColumnsThisLevel)

        reference_list = self.get_list(LevelKey.N,'reference', numColumnsThisLevel, numColumnsNextLevel)


        config.append(perception_list)
        config.append(output_list[0])
        config.append(reference_list)

        return config

    def get_level0top(self, num_inputs, numColumnsThisLevel, num_actions):
        config0=[]

        perception_list = self.get_list(LevelKey.ZERO,'perception', num_inputs, numColumnsThisLevel)

        output_list = self.get_list(LevelKey.ZERO,'output', num_actions, numColumnsThisLevel)

        reference_list = self.get_list(LevelKey.TOP,'reference', 1, numColumnsThisLevel)

        action_list = self.get_list(LevelKey.ZERO,'action', num_actions, numColumnsThisLevel)

        config0.append(perception_list)
        config0.append(output_list[0])
        config0.append(reference_list)
        config0.append(action_list)

        return config0

    def get_leveltop(self, numColumnsThisLevel, numColumnsPreviousLevel, num_top_inputs):
        config=[]

        if num_top_inputs > 0:
            perception_list = self.get_list_assymetric(LevelKey.TOP,'perception', num_top_inputs, numColumnsPreviousLevel, numColumnsThisLevel)
        else:
            perception_list = self.get_list(LevelKey.TOP,'perception', numColumnsPreviousLevel, numColumnsThisLevel)

        output_list = self.get_list(LevelKey.TOP,'output', 1, numColumnsThisLevel)

        reference_list = self.get_list(LevelKey.TOP,'reference', 1, numColumnsThisLevel)


        config.append(perception_list)
        config.append(output_list[0])
        config.append(reference_list)

        return config

    # assume same for all levels and that datatypes are always floats
    """
    def set_output_function(self, node,  thislevel, column, input_weights):
        func = node.get_function('output')
        func.set_name(f'OL{thislevel}C{column}ws')

        weights=[]
        weights.append(input_weights[column])
        func.weights=np.array(weights)


    def set_action_function(self, hpct, env, numColumnsThisLevel,  weights):
        numActions = len(weights)
        for actionIndex in range(numActions):
            action = WeightedSum(weights=weights[actionIndex], name=f'Action{actionIndex+1}ws')
            for column in range(numColumnsThisLevel):
                action.add_link(f'OL0C{column}ws')
            hpct.add_postprocessor(action)
            env.add_link(action)
    """


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/00_examples.ipynb (unless otherwise specified).

__all__ = []

# Cell
from .evolvers import StandardEvolver
from .evolvers import EvolverWrapper
from .configs import BaseConfiguration
from .configs import ProportionalConfiguration
from .configs import ConfigurationStructure
from pct.environments import CartPoleV1
from pct.architectures import ProportionalArchitecture
import random
import warnings
warnings.simplefilter(action='ignore', category=UserWarning)


# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"BaseEvolver": "01_evolvers.ipynb",
         "EvolverWrapper": "01_evolvers.ipynb",
         "SigmoidEvolver": "01_evolvers.ipynb",
         "StandardEvolver": "01_evolvers.ipynb",
         "DynamicEvolver": "01_evolvers.ipynb",
         "check_hash_file_exists": "01_evolvers.ipynb",
         "evolve_from_properties_file": "01_evolvers.ipynb",
         "BaseConfiguration": "02_configs.ipynb",
         "ProportionalConfiguration": "02_configs.ipynb",
         "ConfigurationStructure": "02_configs.ipynb",
         "DynamicConfigurationStructure": "02_configs.ipynb",
         "DynamicConfiguration": "02_configs.ipynb",
         "BinaryOnes": "03_structure.ipynb",
         "BaseParameterType": "03_structure.ipynb",
         "Binary": "03_structure.ipynb",
         "Float": "03_structure.ipynb",
         "Literal": "03_structure.ipynb",
         "Smooth": "03_structure.ipynb",
         "ParameterFactory": "03_structure.ipynb",
         "StructureDefinition": "03_structure.ipynb",
         "generate_evolvers": "04_properties.ipynb",
         "generate_option_files": "04_properties.ipynb",
         "additional_properties": "04_properties.ipynb",
         "write_to_file": "04_properties.ipynb",
         "structure_parameters": "04_properties.ipynb",
         "configurable_parameters": "04_properties.ipynb",
         "description": "04_properties.ipynb",
         "fixed_parameters": "04_properties.ipynb"}

modules = ["examples.py",
           "evolvers.py",
           "configs.py",
           "structure.py",
           "properties.py"]

doc_url = "https://perceptualrobots.github.io/epct/"

git_url = "https://github.com/perceptualrobots/epct/tree/master/"

def custom_doc_links(name): return None
